from enum import Enum

from .utils import PlatFormQuant, Singleton


class LayerFusion(Enum):
    LayerFusionConvBN = 1


class LayerQuantPredictor(Enum):
    LayerPredictorConvRelu = 1
    LayerPredictorBNRelu = 2


class Configure(Singleton):
    class PlatFormQuantConfig():
        platform_quant = PlatFormQuant.luna_quant

    class IQTensorConfig():
        iqmul = True
        iqadd = True
        iqcat = True
        iqclamp = True
        iqsigmoid = True
        iqdiv = True
        iqsum = True
        iqtanh = True
        softmaxint = True
        logsoftmaxint = True
        iqvar = True

    class FunctionConfig():
        linear = True
        bmm = False
        channel_shuffle = False

    class IQCat2AddZeroConfig():
        iqadd2addzero = False

    class BnMomentumUpdateConfig():
        disable = False

    PlatFormQuant = PlatFormQuantConfig()
    IQTensor = IQTensorConfig()
    FunctionQuant = FunctionConfig()
    IQCat2AddZero = IQCat2AddZeroConfig()
    BnMomentumUpdate = BnMomentumUpdateConfig()

config = Configure()

def SetPlatFormQuant(platform_quant=PlatFormQuant.luna_quant):
    r"""设置是否采用luna_quant量化方式

    Args:
        platform_quant(PlatForm.luna_quant):采用何种硬件量化方式，默认luna量化方式

    """
    config.PlatFormQuant.platform_quant = platform_quant



def SetIQTensorMul(enable):
    r"""设置是否启用iqadd功能,默认启用

    Args:
        enable(bool):是否启用

    """
    config.IQTensor.iqmul = enable


def SetIQTensorDiv(enable):
    r"""设置是否启用iqdiv功能,默认启用

    Args:
        enable(bool):是否启用

    """
    config.IQTensor.iqdiv = enable


def SetIQTensorAdd(enable):
    r"""设置是否启用iqadd功能,默认启用

    Args:
        enable(bool):是否启用

    """
    config.IQTensor.iqadd = enable


def SetIQTensorSum(enable):
    r"""设置是否启用iqdiv功能,默认启用

    Args:
        enable(bool):是否启用

    """
    config.IQTensor.iqsum = enable


def SetIQTensorCat(enable):
    r"""设置是否启用iqcat功能,默认启用

    Args:
        enable(bool):是否启用

    """
    config.IQTensor.iqcat = enable


def SetIQTensorSigmoid(enable):
    r"""设置是否启用iqsigmoid功能,默认启用

    Args:
        enable(bool):是否启用

    """
    config.IQTensor.iqsigmoid = enable

def SetIQTensorSoftmax(enable):
    r"""设置是否启用softmaxInt功能,默认启用

    Args:
        enable(bool):是否启用

    """
    config.IQTensor.softmaxint = enable

def SetIQTensorLogSoftmax(enable):
    r"""设置是否启用LogSoftmax功能,默认启用

    Args:
        enable(bool):是否启用

    """
    config.IQTensor.logsoftmaxint = enable

def SetIQTensorTanh(enable):
    r"""设置是否启用iqtanh功能,默认启用

    Args:
        enable(bool):是否启用

    """
    config.IQTensor.iqtanh = enable


def SetIQTensorClamp(enable):
    r"""设置是否启用iqclamp功能,默认启用

    Args:
        enable(bool):是否启用

    """
    config.IQTensor.iqclamp = enable

def SetIQTensorVar(enable):
    r"""设置是否启用iqVar功能,默认启用

    Args:
        enable(bool):是否启用

    """
    config.IQTensor.iqvar = enable


def SetFunctionLinearQuant(enable):
    r"""设置是否启用F.linear量化功能,默认启用

    Args:
        enable(bool):是否启用

    """
    config.FunctionQuant.linear = enable


def SetFunctionBmmQuant(enable):
    r"""设置是否启用torch.bmm量化功能,默认启用

    Args:
        enable(bool):是否启用

    """
    config.FunctionQuant.bmm = enable

def SetFunctionChannelShuffleQuant(enable):
    config.FunctionQuant.channel_shuffle = enable


def SetCastorBiasInt16(bias_int16=True):
    config.CastorBiasInt16.bias_int16 = bias_int16


def SetBnMomentumUpdate(disable=True):
    config.BnMomentumUpdate.disable = disable


def SetIQCat2AddZero(enable=True):
    config.IQCat2AddZero.iqadd2addzero = enable


__all__ = ['config', 'SetPlatFormQuant', 'SetIQTensorAdd', 'SetFunctionLinearQuant', 'SetFunctionBmmQuant', 'SetIQTensorClamp',
           'SetIQTensorCat', 'SetIQTensorSigmoid', 'SetIQTensorSoftmax', 'SetIQTensorLogSoftmax', 'SetIQTensorTanh', 'SetIQTensorDiv', 'SetIQTensorMul', 'SetIQTensorSum',
           'LayerFusion', 'LayerQuantPredictor', 'SetIQCat2AddZero', 'SetBnMomentumUpdate', 'SetIQTensorVar', 'SetFunctionChannelShuffleQuant']



from typing import List

import onnx

from .transform.onnx_utils import remove_identity_node
from .transform.replace_add import replace_add_to_iqadd
from .transform.replace_avgpoolint import replace_avgpool2dint
from .transform.replace_clip import replace_clip_attr
from .transform.replace_conv import replace_conv2dint
from .transform.replace_iqsigmoid import replace_iqsigmoid
from .transform.replace_linearint import replace_linearint

supported_replace_ops = ['Conv', 'ConvTranspose', 'Add', 'Clip', 'AveragePool', 'MaxPool', 'Relu', 'Gemm', 'Transpose', 'Reshape', 'Squeeze', 'Unsqueeze',
                         'Split', "Sigmoid"]


def onnx_quant(model_path, quant_ops_type: List[str], remove_ops_type: List[str] = [], scale_x=16.0, scale_y=16.0, scale_w=64.0, scale_o=16.0, platform_quant="luna_quant"):
    model = onnx.load(model_path)
    for remove_op_type in remove_ops_type:
        model = remove_identity_node(model, remove_op_type)
    for quant_op_type in quant_ops_type:
        if quant_op_type == "Conv" or quant_op_type == "ConvTranspose":
            model = replace_conv2dint(
                model, scale_x, scale_w, scale_o, platform_quant)
        if quant_op_type == "Add":
            model = replace_add_to_iqadd(
                model, scale_x, scale_y, scale_o, platform_quant)
        if quant_op_type == "Clip":
            model = replace_clip_attr(model, scale_o)
        if quant_op_type == "Sigmoid":
            model = replace_iqsigmoid(model, scale_x, 256.0, platform_quant)
        if quant_op_type == "AveragePool":
            model = replace_avgpool2dint(
                model, scale_x, scale_o, platform_quant)
        if quant_op_type == "Gemm":
            model = replace_linearint(
                model, scale_x, scale_w, scale_o, platform_quant)

    # change inp/outp type to int8, ignore quant/dequant op
    for i in range(len(model.graph.input)):
        if model.graph.input[i].type.tensor_type.elem_type == 1:
            model.graph.input[i].type.tensor_type.elem_type = 3
    for i in range(len(model.graph.output)):
        if model.graph.output[i].type.tensor_type.elem_type == 1:
            if "Sigmoid" in quant_ops_type:
                model.graph.output[i].type.tensor_type.elem_type = 2
            else:
                model.graph.output[i].type.tensor_type.elem_type = 3

    onnx.save(model, model_path[:-5]+"_quant.onnx")




import linger
import linger.onnx
import torch
import torch.nn as nn
import numpy
import onnx



# class NetRaw_2(nn.Module):
#     def __init__(self):
#         super(NetRaw_2, self).__init__()
#         self.conv0 = nn.Conv2d(1, 100, kernel_size=(3,1), padding=(0,0), groups=1, bias=True)
#         self.bn0 = nn.BatchNorm2d(100)
#         self.relu0 = nn.ReLU()
#         self.conv1 = nn.Conv2d(100, 100, kernel_size=(1,3), padding=(0,1), groups=1, bias=True)
#         self.bn1 = nn.BatchNorm2d(100)
#         self.relu1 = nn.ReLU()
#         self.final_conv = nn.Conv2d(100, 10, 1, 1, 0)
#     def forward(self, input):
#         x = self.conv0(input)
#         x = self.bn0(x)         
#         x = self.relu0(x)
#         x = self.conv1(x)
#         x = self.bn1(x)
#         x = self.relu1(x)
#         x = self.final_conv(x)  #(1, 10, 1, 100) (d, b*t)
#         return x



# class NetComplicate_2(nn.Module):
#     def __init__(self):
#         super(NetComplicate_2, self).__init__()
#         self.conv0 = nn.Conv2d(1, 100, kernel_size=(3,1), padding=(0,0), groups=1, bias=True)
#         self.bn0 = nn.BatchNorm2d(100)
#         self.relu0 = nn.ReLU()
#         self.conv1 = nn.Conv2d(100, 100, kernel_size=(1,3), padding=(0,1), groups=1, bias=True)
#         self.bn1 = nn.BatchNorm2d(100)
#         self.relu1 = nn.ReLU()
#         self.final_conv = nn.Conv2d(100, 10, 1, 1, 0)
#         self.net_2 = NetRaw_2()
#     def forward(self, input):
#         x = self.conv0(input)
#         y = self.net_2(input)
#         x = self.bn0(x)         
#         x = self.relu0(x)
#         x = self.conv1(x)
#         x = self.bn1(x)
#         x = self.relu1(x)
#         x = self.final_conv(x)  #(1, 10, 1, 100) (d, b*t)
#         return x+y
# def test_scope_info_simple():            
#     torch.manual_seed(2)
#     torch.cuda.manual_seed_all(2)
#     aa = torch.randn(10, 1, 10, 10).cuda()
#     netraw_2 = NetRaw_2().cuda()      
#     with torch.no_grad():
#         netraw_2.eval()
#         linger.onnx.export(netraw_2, (aa), "data.ignore/test_scoped_1.onnx",export_params=True,opset_version=12, operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
#     onnx_model = onnx.load("data.ignore/test_scoped_1.onnx")
#     for node in onnx_model.graph.node:
#        node_name =node.name
#       # print(node_name)
#        assert node_name.startswith('.')
#        assert len(node_name.split('.')) >1
#        assert len(node_name.split('/')) ==2

# def test_scope_info_complicate():            
#     torch.manual_seed(2)
#     torch.cuda.manual_seed_all(2)
#     aa = torch.randn(10, 1, 10, 10).cuda()
#     netcomplicate_2 = NetComplicate_2().cuda()      
#     with torch.no_grad():
#         netcomplicate_2.eval()
#         linger.onnx.export(netcomplicate_2, (aa), "data.ignore/test_scoped_2.onnx",export_params=True,opset_version=12, operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
#     onnx_model = onnx.load("data.ignore/test_scoped_2.onnx")
#     for node in onnx_model.graph.node:
#        node_name =node.name
#        #print(node_name)
#        assert node_name.startswith('.')
#        assert len(node_name.split('.')) >1
#        if node.op_type !="Add":
#            assert len(node_name.split('/')) ==2
#        assert node.op_type != 'ScopedEnter'
#        assert node.op_type != 'ScopedLeave'




import torch
import torch.nn as nn
import torch.nn.functional as F

from ..quant import NormalizeFunction


class NormalizeLinear(nn.Linear):
    def __init__(self, in_features: int, out_features: int, bias: bool = True,
                 normalize_data=None, normalize_weight=None, normalize_bias=None, ahead_relu=False, ahead_sigmoid=False) -> None:
        assert normalize_data is None or normalize_data > 0, 'normalize value is None or must >0'
        assert normalize_weight is None or normalize_weight > 0, 'normalize value is None or must >0'
        assert normalize_bias is None or normalize_bias > 0, 'normalize value is None or must >0'
        nn.Linear.__init__(self, in_features, out_features, bias)
        self.normalize_data = normalize_data
        self.normalize_weight = normalize_weight
        self.normalize_bias = normalize_bias
        self.ahead_relu = ahead_relu
        self.ahead_sigmoid = ahead_sigmoid

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        normalized_weight = self.weight
        if self.normalize_weight is not None:
            normalized_weight = NormalizeFunction.apply(
                normalized_weight, self.normalize_weight, self.training)
        normalized_bias = self.bias
        if (self.bias is not None) and (self.normalize_bias is not None):
            normalized_bias = NormalizeFunction.apply(
                normalized_bias, self.normalize_bias, self.training)
        out = F.linear(input, normalized_weight, normalized_bias)
        if self.normalize_data:
            out = NormalizeFunction.apply(out, self.normalize_data, self.training, False)
        return out

    def extra_repr(self):
        s = nn.Linear.extra_repr(self)
        extra_s = ',normalize_data:{normalize_data},normalize_weight:{normalize_weight},normalize_bias:{normalize_bias},ahead_relu:{ahead_relu}'.format(
            **self.__dict__)
        return s+extra_s


__all__ = ['NormalizeLinear']



import os

import linger
import numpy as np
import torch
import torch.nn as nn

if not os.path.exists("data.ignore"):
    os.makedirs("data.ignore")


def test_avgpool2dint():
    class Net(nn.Module):
        def __init__(self):
            super(Net, self).__init__()
            self.conv = nn.Conv2d(10, 10, kernel_size=3, stride=1,
                                  padding=1, bias=True)
            self.bn = nn.BatchNorm2d(10)
            self.relu = nn.ReLU()
            self.pool = nn.AvgPool2d((2, 2), (2, 2), (0, 0), False)
            self.fc = nn.Linear(250, 100)

        def forward(self, x):
            x = self.conv(x)
            x = self.bn(x)
            x = self.relu(x)
            n, c, h, w = x.shape
            x = self.pool(x)
            n, c, h, w = x.shape
            x = x.view((n, c*h*w))
            x = self.fc(x)
            return x

    torch.manual_seed(1)
    torch.cuda.manual_seed_all(1)
    np.random.seed(1)

    torch.cuda.set_device(0)
    net = Net().cuda()
    dummy_input = torch.randn(1, 10, 10, 10).cuda()
    target = torch.ones(1, 100).cuda()
    criterion = nn.MSELoss()
    replace_tuple = (nn.Conv2d, nn.Linear, nn.AvgPool2d)
    net.train()
    linger.SetPlatFormQuant(platform_quant=linger.PlatFormQuant.luna_quant)
    # linger.disable_quant(net.fc)
    net = linger.init(net, quant_modules=replace_tuple,
                      mode=linger.QuantMode.QValue)
    optimizer = torch.optim.SGD(net.parameters(), lr=0.01)
    loss = None
    for i in range(200):
        optimizer.zero_grad()
        out = net(dummy_input)
        loss = criterion(out, target)
        if i % 20 == 0:
            print('loss: ', loss)
        # if i == 190:
        #     import pdb; pdb.set_trace()
        loss.backward()
        optimizer.step()
    net.eval()
    torch.save(net.state_dict(), 'data.ignore/aa.pt')

    out1 = net(dummy_input)
    with torch.no_grad():
        torch.onnx.export(net, dummy_input, "data.ignore/avg_pool.onnx", export_params=True, opset_version=11,
                          operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    print("out1.mean(): ", out1.mean())
    assert abs(out1.mean() - 1.0) < 0.15



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import torch
import torch.nn as nn
import linger
import numpy as np

def test_conv_linear():
    class Net(nn.Module):
        def __init__(self):
            super(Net, self).__init__()
            self.conv = nn.Conv2d(10, 10, kernel_size=3, stride=1,
                         padding=1, bias=True)
            self.bn = nn.BatchNorm2d(10)
            self.relu = nn.ReLU()
            self.conv1 = nn.Conv2d(10, 10, kernel_size=3, stride=1,
                         padding=1, bias=True)
            self.conv2 = nn.Conv2d(10, 10, kernel_size=3, stride=1,
                         padding=1, bias=True)  
            self.bn1 = nn.BatchNorm2d(10)
            self.relu1 = nn.ReLU()           
            self.fc = nn.Linear(1000, 100)

        def forward(self, x):
            x = self.conv(x)
            x = self.bn(x)
            x = self.relu(x)
            x = x - 1
            x = self.conv1(x)
            x = self.conv2(x)
            x = self.bn1(x)
            x = self.relu1(x)
            n, c, h, w = x.shape
            x = x.view((n, c*h*w))
            x = self.fc(x)
            return x
    torch.manual_seed(1)
    torch.cuda.manual_seed_all(1)
    np.random.seed(1)
    # random.seed(1)

    torch.cuda.set_device(0)
    net = Net().cuda()
    aa = torch.randn(1, 10, 10, 10).cuda()
    target = torch.ones(1, 100).cuda()
    criterion = nn.MSELoss()
    replace_tuple=(nn.Conv2d, nn.Linear)
    net.train()
    net = linger.normalize_layers(net, normalize_modules=replace_tuple, normalize_weight_value=4, normalize_bias_value=4, normalize_output_value=4)
    net.cuda()
    optimizer = torch.optim.SGD(net.parameters(), lr = 0.01)
    loss = None
    for i in range(200):
        optimizer.zero_grad()
        out = net(aa)
        loss = criterion(out, target)
        if i % 20 == 0:
            print('loss: ', loss)
        loss.backward()
        optimizer.step()
    # assert loss < 1e-12, 'training loss error'
    net.eval()
    torch.save(net.state_dict(), 'data.ignore/conv_linear.pt')
    out1 = net(aa)
    #print(out1)
    with torch.no_grad():
        torch.onnx.export(net, aa,"data.ignore/conv_linear.onnx",export_params=True,opset_version=11,operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    assert abs(out1.mean() - 1) < 0.01


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from .normalize_layernorm import NormalizeLayerNorm
from .normalize_fastLSTM import NormalizeFastLSTM
from .normalize_fastGRU import NormalizeFastGRU
import torch.nn as nn

from .normalize_conv1d import NormalizeConv1d
from .normalize_conv2d import NormalizeConv2d
from .normalize_convbn1d import NormalizeConvBN1d
from .normalize_convbn2d import NormalizeConvBN2d
from .normalize_convTranspose2d import NormalizeConvTranspose2d
from .normalize_linear import NormalizeLinear

from.normalize_batchnorm2d import NormalizeBatchNorm2d


DefaultNormalizeIntXModule = (nn.Conv2d, nn.Linear, nn.ConvTranspose2d, NormalizeConvBN2d,
                              nn.Conv1d, NormalizeConvBN1d, nn.BatchNorm2d, nn.GRU, nn.LSTM)
SupportNormalizeTorchModules = [nn.Conv2d, nn.Linear, nn.ConvTranspose2d, NormalizeConvBN2d,
                                nn.Conv1d, NormalizeConvBN1d, nn.BatchNorm2d, nn.GRU, nn.LSTM, nn.Embedding]
SupportNormalizeIntXModules = (NormalizeConv2d, NormalizeLinear, NormalizeConvTranspose2d, NormalizeConvBN2d,
                               NormalizeConv1d, NormalizeConvBN1d, NormalizeBatchNorm2d, NormalizeFastGRU, NormalizeFastLSTM, NormalizeLayerNorm)



import torch
from linger import NormalizeConv2d


def test_normalize_conv2d_foward():
    module = NormalizeConv2d(2, 4, (3, 3), normalize_data=4,
                             normalize_weight=4, normalize_bias=4)
    module.weight.data.fill_(8)
    module.bias.data.fill_(8)
    input = 8 * torch.randn(1, 2, 4, 4)
    assert (input < 8).any()
    assert (input > 4).any()
    m = module(input)
    assert (m < 4.0001).all()


#!/usr/bin/env python
# -*- encoding: utf-8 -*-

import torch
from linger import NormalizeConvTranspose2d


def test_normalize_linear_foward():
    module = NormalizeConvTranspose2d(
        128, 512, 3, bias=True, normalize_data=4, normalize_weight=4, normalize_bias=4)
    module.weight.data.fill_(8)
    module.bias.data.fill_(8)
    input = 8 * torch.randn(2, 128, 10, 10)
    assert (input < 8).any()
    assert (input > 4).any()
    m = module(input)
    assert (m < 4.0001).all()



import os

import linger
import torch
from linger.ops import (IQTensor, from_torch_tensor, iqadd, iqAddLayer,
                             iqmul, iqMulLayer, torch_cat)

if not os.path.exists('data.ignore'):
    os.mkdir('data.ignore')


def test_iqMul():
    x = torch.tensor([[6.0, 4.0]], device=torch.device(
        'cuda:0'), requires_grad=True)

    y = torch.tensor([[3.0, 8.0]], device=torch.device(
        'cuda:0'), requires_grad=True)
    iq_layer = iqMulLayer().cuda()
    a = from_torch_tensor(x, 127.0/6.0, 8)
    b = from_torch_tensor(y, 127.0/8, 8)

    m = iq_layer(a, b, 127.0/32)  # iqAdd.apply(x,y).sum().backward()
    assert abs(m.data[0][0].item()-18.1417) < 0.2
    assert abs(m.data[0][1].item()-32) < 0.3
    m.sum().backward()

    p = torch.tensor([[6.0, 4.0]], device=torch.device(
        'cuda:0'), requires_grad=True).cuda()
    q = torch.tensor([[3.0, 8.0]], device=torch.device(
        'cuda:0'), requires_grad=True).cuda()
    m = p*q
    m = m.sum()
    m.backward()
    err = p.grad.data - x.grad.data
    assert err.abs().sum() < 0.01
    err = q.grad.data - y.grad.data
    assert err.abs().sum() < 0.01


def test_iqMul_layer():
    x = torch.tensor([[6.0, 4.0]], device=torch.device(
        'cuda:0'), requires_grad=True)
    y = torch.tensor([[3.0, 8.0]], device=torch.device(
        'cuda:0'), requires_grad=True)

    class iqTestLayer(torch.nn.Module):
        def __init__(self):
            super(iqTestLayer, self).__init__()

        def forward(self, x, y):
            return iqmul(self, x, y, 'test')

    a = from_torch_tensor(x, 127.0/6.0, 8)
    b = from_torch_tensor(y, 127.0/8, 8)

    net = iqTestLayer().cuda()  # iqAdd.apply(x,y).sum().backward()
    m = net(a, b)
    assert abs(m.data[0][0].item()-18.1417) < 0.2
    assert abs(m.data[0][1].item()-32) < 0.3
    m.sum().backward()

    p = torch.tensor([[6.0, 4.0]], device=torch.device(
        'cuda:0'), requires_grad=True)
    q = torch.tensor([[3.0, 8.0]], device=torch.device(
        'cuda:0'), requires_grad=True)
    m = p*q
    m = m.sum()
    m.backward()
    err = p.grad.data - x.grad.data
    assert a.grad is None
    assert err.abs().sum() < 0.01
    err = q.grad.data - y.grad.data
    assert b.grad is None
    assert err.abs().sum() < 0.01


def test_iqMul_layer_cpu():
    x = torch.tensor([[6.0, 4.0]], requires_grad=True)
    y = torch.tensor([[3.0, 8.0]], requires_grad=True)

    class iqTestLayer(torch.nn.Module):
        def __init__(self):
            super(iqTestLayer, self).__init__()

        def forward(self, x, y):
            return iqmul(self, x, y, 'test')

    a = from_torch_tensor(x, 127.0/6.0, 8)
    b = from_torch_tensor(y, 127.0/8, 8)

    net = iqTestLayer()
    m = net(a, b)
    # print(m)
    assert abs(m.data[0][0].item()-18.1417) < 0.2
    assert abs(m.data[0][1].item()-32) < 0.3
    m.sum().backward()

    p = torch.tensor([[6.0, 4.0]], requires_grad=True)
    q = torch.tensor([[3.0, 8.0]], requires_grad=True)
    m = p*q
    m = m.sum()
    m.backward()
    err = p.grad.data - x.grad.data
    assert a.grad is None
    assert err.abs().sum() < 0.01
    err = q.grad.data - y.grad.data
    assert b.grad is None
    assert err.abs().sum() < 0.01


def test_iqmul_module():
    class TestModel(torch.nn.Module):
        def __init__(self):
            super(TestModel, self).__init__()
            self.aa = torch.nn.Parameter(torch.zeros((1)))

        def forward(self, x, y):
            return x * y
    net = TestModel().cuda()
    x = torch.tensor([[6.0, 4.0]], requires_grad=True)
    y = torch.tensor([[3.0, 8.0]], requires_grad=True)
    a = from_torch_tensor(x, 127.0/6.0, 8)
    b = from_torch_tensor(y, 127.0/8, 8)
    net = linger.init(net)
    z = net(a, b)
    assert isinstance(z, IQTensor)


def test_iqimul_module():
    class TestModel(torch.nn.Module):
        def __init__(self):
            super(TestModel, self).__init__()
            self.aa = torch.nn.Parameter(torch.zeros((1)))

        def forward(self, x, y):
            x *= y
            return x
    net = TestModel().cuda()
    x = torch.tensor([[6.0, 4.0]], requires_grad=True)
    y = torch.tensor([[3.0, 8.0]], requires_grad=True)
    a = from_torch_tensor(x, 127.0/6.0, 8)
    b = from_torch_tensor(y, 127.0/8, 8)
    net = linger.init(net)
    z = net(a, b)
    assert isinstance(z, IQTensor)


def test_iqAdd():
    x = torch.tensor([[6.0, 4.0]], device=torch.device(
        'cuda:0'), requires_grad=True)

    y = torch.tensor([[3.0, 8.0]], device=torch.device(
        'cuda:0'), requires_grad=True)
    iq_layer = iqAddLayer().cuda()
    a = from_torch_tensor(x, 127.0/6.0, 8)
    b = from_torch_tensor(y, 127.0/8, 8)

    m = iq_layer(a, b, 127.0/12)  # iqAdd.apply(x,y).sum().backward()
    assert m.data[0][0]-16.535 < 0.1
    assert m.data[0][1]-20.031 < 0.1
    m.sum().backward()

    p = torch.tensor([[6.0, 4.0]], device=torch.device(
        'cuda:0'), requires_grad=True).cuda()
    q = torch.tensor([[3.0, 8.0]], device=torch.device(
        'cuda:0'), requires_grad=True).cuda()
    m = p+q
    m = m.sum()
    m.backward()
    err = p.grad.data - x.grad.data
    assert err.abs().sum() < 0.01
    err = q.grad.data - y.grad.data
    assert err.abs().sum() < 0.01

def test_iqAdd_layer():
    x = torch.tensor([[6.0, 4.0]], device=torch.device(
        'cuda:0'), requires_grad=True)
    y = torch.tensor([[3.0, 8.0]], device=torch.device(
        'cuda:0'), requires_grad=True)

    class iqTestLayer(torch.nn.Module):
        def __init__(self):
            super(iqTestLayer, self).__init__()

        def forward(self, x, y):
            return iqadd(self, x, y, 'test')

    a = from_torch_tensor(x, 127.0/6.0, 8)
    b = from_torch_tensor(y, 127.0/8, 8)

    net = iqTestLayer().cuda()  # iqAdd.apply(x,y).sum().backward()
    m = net(a, b)
    assert m.data[0][0]-16.535 < 0.1
    assert m.data[0][1]-20.031 < 0.1
    m.sum().backward()

    p = torch.tensor([[6.0, 4.0]], device=torch.device(
        'cuda:0'), requires_grad=True)
    q = torch.tensor([[3.0, 8.0]], device=torch.device(
        'cuda:0'), requires_grad=True)
    m = p+q
    m = m.sum()
    m.backward()
    err = p.grad.data - x.grad.data
    assert a.grad is None
    assert err.abs().sum() < 0.01
    err = q.grad.data - y.grad.data
    assert b.grad is None
    assert err.abs().sum() < 0.01


def test_iqAdd_layer_cpu():
    x = torch.tensor([[6.0, 4.0]], requires_grad=True)
    y = torch.tensor([[3.0, 8.0]], requires_grad=True)

    class iqTestLayer(torch.nn.Module):
        def __init__(self):
            super(iqTestLayer, self).__init__()

        def forward(self, x, y):
            return iqadd(self, x, y, 'test')

    a = from_torch_tensor(x, 127.0/6.0, 8)
    b = from_torch_tensor(y, 127.0/8, 8)

    net = iqTestLayer()
    m = net(a, b)
    assert m.data[0][0]-16.535 < 0.1
    assert m.data[0][1]-20.031 < 0.1
    m.sum().backward()

    p = torch.tensor([[6.0, 4.0]], requires_grad=True)
    q = torch.tensor([[3.0, 8.0]], requires_grad=True)
    m = p+q
    m = m.sum()
    m.backward()
    err = p.grad.data - x.grad.data
    assert a.grad is None
    assert err.abs().sum() < 0.01
    err = q.grad.data - y.grad.data
    assert b.grad is None
    assert err.abs().sum() < 0.01


def test_iqadd_module():
    class TestModel(torch.nn.Module):
        def __init__(self):
            super(TestModel, self).__init__()
            self.aa = torch.nn.Parameter(torch.zeros((1)))

        def forward(self, x, y):
            return x + y
    net = TestModel().cuda()
    x = torch.tensor([[6.0, 4.0]], requires_grad=True)
    y = torch.tensor([[3.0, 8.0]], requires_grad=True)
    a = from_torch_tensor(x, 127.0/6.0, 8)
    b = from_torch_tensor(y, 127.0/8, 8)
    net = linger.init(net)
    z = net(a, b)
    assert isinstance(z, IQTensor)


def test_iqiadd_module():
    class TestModel(torch.nn.Module):
        def __init__(self):
            super(TestModel, self).__init__()
            self.aa = torch.nn.Parameter(torch.zeros((1)))

        def forward(self, x, y):
            x += y
            return x
    net = TestModel().cuda()
    x = torch.tensor([[6.0, 4.0]], requires_grad=True)
    y = torch.tensor([[3.0, 8.0]], requires_grad=True)
    a = from_torch_tensor(x, 127.0/6.0, 8)
    b = from_torch_tensor(y, 127.0/8, 8)
    net = linger.init(net)
    z = net(a, b)
    assert isinstance(z, IQTensor)


def test_iqcat_function():
    class TestModel(torch.nn.Module):
        def __init__(self):
            super(TestModel, self).__init__()
            self.aa = torch.nn.Parameter(torch.zeros((1)))

        def forward(self, x, y, z):
            return torch.cat((x, y, z), dim=0)

    x = torch.tensor([[6.0, 4.0]], requires_grad=True)
    y = torch.tensor([[3.0, 8.0]], requires_grad=True)
    z = torch.tensor([[5.0, 7.0]], requires_grad=True)

    a = torch_cat((x, y, z), dim=0)

    a.sum().backward()

    x_ = torch.tensor([[6.0, 4.0]], requires_grad=True)
    y_ = torch.tensor([[3.0, 8.0]], requires_grad=True)
    z_ = torch.tensor([[5.0, 7.0]], requires_grad=True)
    x_iq = from_torch_tensor(x_, 127.0/6.0, 8)
    y_iq = from_torch_tensor(y_, 127.0/8.0, 8)
    z_iq = from_torch_tensor(z_, 127.0/7.0, 8)

    net = TestModel()
    net = linger.init(net)
    b = net(x_iq, y_iq, z_iq)
    b.sum().backward()
    assert (b-a).sum() < 0.2
    err = x_.grad.data - x.grad.data
    assert err.sum() < 0.1
    err = y_.grad.data - y.grad.data
    assert err.sum() < 0.1
    err = z_.grad.data - z.grad.data
    assert err.sum() < 0.1


def test_iqcat_running_o():
    class TestModel(torch.nn.Module):
        def __init__(self):
            super(TestModel, self).__init__()
            self.aa = torch.nn.Parameter(torch.zeros((1)))

        def forward(self, x, y, z):
            return torch.cat((x, y, z), dim=0)

    x_ = torch.tensor([[6.0, 4.0]], requires_grad=True)
    y_ = torch.tensor([[3.0, 8.0]], requires_grad=True)
    z_ = torch.tensor([[5.0, 7.0]], requires_grad=True)
    x_iq = from_torch_tensor(x_, 127.0/6.0, 8)
    y_iq = from_torch_tensor(y_, 127.0/8.0, 8)
    z_iq = from_torch_tensor(z_, 127.0/7.0, 8)

    net = TestModel()
    net = linger.init(net)
    b = None
    for _ in range(100):
        b = net(x_iq, y_iq, z_iq)
    net.eval()
    c = net(x_iq, y_iq, z_iq)
    assert (c - b).abs().sum() < 0.1


def test_iqcat_state_dict():
    class TestModel(torch.nn.Module):
        def __init__(self):
            super(TestModel, self).__init__()
            self.aa = torch.nn.Parameter(torch.zeros((1)))

        def forward(self, x, y, z):
            return torch.cat((x, y, z), dim=0)

    x_ = torch.tensor([[6.0, 4.0]], requires_grad=True)
    y_ = torch.tensor([[3.0, 8.0]], requires_grad=True)
    z_ = torch.tensor([[5.0, 7.0]], requires_grad=True)
    x_iq = from_torch_tensor(x_, 127.0/6.0, 8)
    y_iq = from_torch_tensor(y_, 127.0/8.0, 8)
    z_iq = from_torch_tensor(z_, 127.0/7.0, 8)

    net = TestModel()
    net = linger.init(net)
    b = None
    for _ in range(100):
        b = net(x_iq, y_iq, z_iq)
    torch.save(net.state_dict(), 'data.ignore/param.dict')

    net2 = TestModel()
    net2 = linger.init(net2)
    net2.load_state_dict(torch.load(
        'data.ignore/param.dict', map_location='cpu'))

    net.eval()
    c = net(x_iq, y_iq, z_iq)
    assert (c - b).abs().sum() < 0.1



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import torch
import torch.nn as nn

from ..quant import NormalizeFunction


class NormalizeBatchNorm2d(nn.BatchNorm2d):
    def __init__(self, num_features: int, eps=1e-5, momentum=0.1, affine=True, track_running_stats=True,
                 normalize_data=None, normalize_weight=None, normalize_bias=None, ahead_relu=False) -> None:
        assert normalize_data is None or normalize_data > 0, 'normalize value is None or must >0'
        assert normalize_weight is None or normalize_weight > 0, 'normalize value is None or must >0'
        assert normalize_bias is None or normalize_bias > 0, 'normalize value is None or must >0'
        nn.BatchNorm2d.__init__(self, num_features, eps,
                                momentum, affine, track_running_stats)
        self.normalize_data = normalize_data
        self.normalize_weight = normalize_weight
        self.normalize_bias = normalize_bias
        self.ahead_relu = ahead_relu

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        batchsize, channels, height, width = input.shape
        size = batchsize * height * width
        if self.training:
            mean = input.sum((0, 2, 3), keepdim=True) / size
            var = input.pow(2).sum((0, 2, 3), keepdim=True) / size - \
                (input.sum((0, 2, 3), keepdim=True) / size).pow(2)
            var = torch.clamp(var, min=0.0)
            self.running_mean = (
                1 - self.momentum) * self.running_mean + self.momentum * mean.squeeze().detach()
            self.running_var = (1 - self.momentum) * self.running_var + \
                self.momentum * var.squeeze().detach()
        else:
            mean = self.running_mean.reshape(1, -1, 1, 1)
            var = self.running_var.reshape(1, -1, 1, 1)
        sigma = 1 / torch.sqrt(var + self.eps)
        alpha = self.weight.view(1, -1, 1, 1) * sigma
        beta = self.bias.view(1, -1, 1, 1) - mean * alpha
        if self.normalize_weight is not None:
            alpha = NormalizeFunction.apply(
                alpha, self.normalize_weight, self.training)
        if self.normalize_bias is not None:
            beta = NormalizeFunction.apply(
                beta, self.normalize_bias, self.training)
        out = alpha * input + beta
        if self.normalize_data is not None:
            out = NormalizeFunction.apply(
                out, self.normalize_data, self.training, False)
        return out

    def extra_repr(self) -> str:
        return 'normalize_data:{normalize_data},normalize_weight:{normalize_weight},normalize_bias:{normalize_bias},ahead_relu:{ahead_relu}'.format(**self.__dict__)


__all__ = ['NormalizeBatchNorm2d']



import torch
import torch.nn as nn
import linger
from linger.ops import from_torch_tensor
from linger.ops import (IQTensor, from_torch_tensor, iqadd, iqAddLayer,
                             iqmul, iqMulLayer, torch_cat)

def test_iqmul_iqtensor_scalar():
    class iqTestLayer(torch.nn.Module):
        def __init__(self):
            super(iqTestLayer, self).__init__()
            self.fc = nn.Linear(2, 1)

        def forward(self, x):
            x = self.fc(x)
            x = x * 0.125
            return x

    model = iqTestLayer().cuda()
    model = linger.init(model, quant_modules=(nn.Linear), parameter_bits=8)

    x = torch.tensor([[0.6, 0.4]], requires_grad=True).cuda()
    # import pdb; pdb.set_trace()
    a = from_torch_tensor(x, 127.0/8, 8)
    
    out = model(x)
    model.eval()

    with torch.no_grad():
        torch.onnx.export(model, (x),"./data.ignore/iqmul.onnx",export_params=True,opset_version=11,operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)

def test_iqmul_scale_x_y_o():
    
    class iqTestLayer(torch.nn.Module):
        def __init__(self):
            super(iqTestLayer, self).__init__()
            self.fc = nn.Linear(2, 1)

        def forward(self, x):
            x = iqmul(self, x, 0.125)
            return x

    x = torch.tensor([[0.6, 0.4]], requires_grad=True).cuda()
    a = from_torch_tensor(x, 16, 8)

    model = iqTestLayer().cuda()
    # model = linger.init(model, quant_modules=(), parameter_bits=8)

    out = model(a)
    model.eval()

    with torch.no_grad():
        torch.onnx.export(model, (a),"./data.ignore/iqmul.onnx",export_params=True,opset_version=11,operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)



import onnx


def replace_avgpool2dint(model, scale_x, scale_o, platform_quant):
    for node in model.graph.node[::-1]:
        if node.op_type == "AveragePool":
            node.op_type = "AvgPool2dInt"
            node.domain = "thinker"
            node.attribute.extend(
                    [onnx.helper.make_attribute("scale_x", scale_x), onnx.helper.make_attribute("scale_o", scale_o), \
                onnx.helper.make_attribute("data_bits", 8), onnx.helper.make_attribute("o_bits", 8), onnx.helper.make_attribute("parameter_bits", 8), \
                onnx.helper.make_attribute("platform_quant", platform_quant) ])
    return model


import torch
import torch.nn

from .modules import NormalizeConvBN1d, NormalizeConvBN2d
from .ops.ops_names import LINGER_AHEAD_RELU, LINGER_AHEAD_SIGMOID, LINGER_IGNORE_PAMAMTER
from .utils import Singleton, get_device


class FuseableConvBN():
    def __init__(self, conv_f, conv, bn_f, bn, root_model=None):
        self.conv_f = conv_f
        self.conv = conv
        self.bn_f = bn_f
        self.bn = bn
        self.scope_conv = None
        self.scope_bn = None
        self.root_model = None

    def set_root_model(self, root_model):
        self.root_model = root_model


class EmptyBatchNorm(torch.nn.Module):
    r"""融合后的BNmoudule占位符,没有进行任何Tensor操作

    """

    def __init__(self):
        super(EmptyBatchNorm, self).__init__()
        setattr(self, LINGER_IGNORE_PAMAMTER,
                torch.nn.Parameter(torch.zeros([1])))

    def forward(self, input):
        return input

    def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs):
        pass


def fuse_conv_bn(conv, bn):
    eps = 1e-5
    c_b = getattr(conv, 'bias', None)

    b_mean = bn.running_mean.data
    b_var = bn.running_var.data
    b_w = bn.weight.data
    b_b = bn.bias.data
    sigma = 1/torch.sqrt(b_var+eps)
    alpha = b_w * sigma
    beta = b_b - b_mean * alpha
    conv.weight.data.mul_(alpha.view(-1, *([1]*(len(conv.weight.shape)-1))))
    if c_b is not None:
        conv.bias.data.mul_(alpha).add_(beta)
    else:
        conv.bias = bn.bias
        conv.bias.data.mul_(0).add_(beta)


class SingletonConvFusedBnModules(Singleton):
    fused_conv_module = {}
    fused_bn_module = {}
    _is_close_register = False

    def _close_register(self):
        self._is_close_register = True

    def _register(self, fuseable_conv_bn):
        if self._is_close_register:
            print("warning: module has initlized and linger.init may not work")
        self.fused_conv_module[fuseable_conv_bn.conv] = fuseable_conv_bn
        self.fused_bn_module[fuseable_conv_bn.bn] = fuseable_conv_bn

    def _is_registered_conv(self, conv):
        f_conv = self.fused_conv_module.get(conv)
        return f_conv

    def _is_registered_bn(self, bn):
        f_bn = self.fused_bn_module.get(bn)
        return f_bn

    def build_normalize_convbn2d_scope(self, model):
        queue = [('', '', model)]
        while len(queue) > 0:
            (node_name, scope_name, node) = queue.pop(0)
            find_fused_info = self._is_registered_conv(node)
            if find_fused_info is not None:
                find_fused_info.scope_conv = scope_name
                if find_fused_info.root_model is None:
                    find_fused_info.set_root_model(model)
                    conv_m = find_fused_info.conv
                    bn_m = find_fused_info.bn
                    conv_have_bias = False if conv_m.bias is None else True
                    clamp_conv = None
                    device = get_device(conv_m)
                    ahead_relu = getattr(
                        conv_m, LINGER_AHEAD_RELU, False)
                    if type(conv_m) == torch.nn.Conv2d:
                        clamp_conv = NormalizeConvBN2d(in_channels=conv_m.in_channels,
                                                       out_channels=conv_m.out_channels, kernel_size=conv_m.kernel_size, stride=conv_m.stride,
                                                       padding=conv_m.padding, dilation=conv_m.dilation, groups=conv_m.groups, bias=conv_have_bias, padding_mode=conv_m.padding_mode,
                                                       eps=bn_m.eps, momentum=bn_m.momentum, affine=bn_m.affine, track_running_stats=bn_m.track_running_stats,
                                                       normalize_data=None, normalize_weight=None, normalize_bias=None, ahead_relu=ahead_relu)
                    elif type(conv_m) == torch.nn.Conv1d:
                        clamp_conv = NormalizeConvBN1d(in_channels=conv_m.in_channels,
                                                       out_channels=conv_m.out_channels, kernel_size=conv_m.kernel_size, stride=conv_m.stride,
                                                       padding=conv_m.padding, dilation=conv_m.dilation, groups=conv_m.groups, bias=conv_have_bias, padding_mode=conv_m.padding_mode,
                                                       eps=bn_m.eps, momentum=bn_m.momentum, affine=bn_m.affine, track_running_stats=bn_m.track_running_stats,
                                                       normalize_data=None, normalize_weight=None, normalize_bias=None, ahead_relu=ahead_relu)
                    clamp_conv = clamp_conv.to(device)
                    setattr(find_fused_info.conv_f, node_name, clamp_conv)
                else:
                    assert find_fused_info.root_model == model
            for name, submodule in node.named_children():
                prefix = '' if scope_name == '' else scope_name+'.'
                queue.append((name, prefix+name, submodule))

    def build_empty_bn_scope(self, model):
        queue = [('', '', model)]
        while len(queue) > 0:
            (node_name, scope_name, node) = queue.pop(0)
            find_fused_info = self._is_registered_bn(node)
            if find_fused_info is not None:
                find_fused_info.scope_bn = scope_name
                if find_fused_info.root_model is None:
                    find_fused_info.set_root_model(model)
                else:
                    assert find_fused_info.root_model == model
                setattr(find_fused_info.bn_f, node_name, EmptyBatchNorm())
            for name, submodule in node.named_children():
                prefix = '' if scope_name == '' else scope_name+'.'
                queue.append((name, prefix+name, submodule))

    @staticmethod
    def get_module(model, scope):
        attr_arr = scope.split('.')
        cur_module = model
        for att in attr_arr:
            cur_module = getattr(cur_module, att, None)
        return cur_module

    def fuse_state_dicts(self, state_dict):
        for v in self.fused_conv_module.values():
            assert v.scope_conv is not None
            assert v.scope_bn is not None

            class GeneralModule():
                pass
            keys_bn = []
            atts_bn = {}
            for key_dict in state_dict.keys():
                prefix = v.scope_bn+'.'
                if key_dict.startswith(prefix):
                    attr_name = key_dict[len(prefix):]
                    attr_name = attr_name.split('.', 1)[0]
                    keys_bn.append(key_dict)
                    atts_bn[attr_name] = state_dict[key_dict]
            keys_conv = []
            atts_conv = {}
            for key_dict in state_dict.keys():
                prefix = v.scope_conv+'.'
                if key_dict.startswith(prefix):
                    attr_name = key_dict[len(prefix):]
                    attr_name = attr_name.split('.', 1)[0]
                    atts_conv[attr_name] = state_dict[key_dict]
                    keys_conv.append(key_dict)
            if LINGER_IGNORE_PAMAMTER not in atts_bn.keys():
                for att, att_dict in atts_conv.items():
                    state_dict[v.scope_conv+'.conv.'+att] = att_dict
                for att, att_dict in atts_bn.items():
                    state_dict[v.scope_conv+'.bn.'+att] = att_dict
                for key_bn_pop in keys_bn:
                    if key_bn_pop != LINGER_IGNORE_PAMAMTER:
                        state_dict.pop(key_bn_pop)
                for key_conv_pop in keys_conv:
                    state_dict.pop(key_conv_pop)

    def clear(self):
        if self._is_close_register:
            print("warning: module has initlized and linger.clear may not work")
        self.fused_conv_module.clear()
        self.fused_bn_module.clear()

    def has_fuseable_items(self):
        return len(self.fused_conv_module) > 0


class OpNodeInfo():
    def __init__(self):
        self.inputs = []
        self.outputs = []
        self.op = None
        self.scope = None

    def __str__(self):
        s = 'input:'
        for i in self.inputs:
            s += i+' '
        s += '\t output:'
        for o in self.outputs:
            s += o+' '
        s += '\t operator:' + self.op
        s += '\t scope:' + self.scope
        return s

    @staticmethod
    def parse_scope_to_path(scope_str):
        tail_name = scope_str.strip().split('/')[-1].strip()
        assert tail_name != ''
        tail_name = tail_name.replace('__module.', '', 1)
        return tail_name

    def parse_scope(self):
        return self.parse_scope_to_path(self.scope)


def _check_version(version_str, major, subor, minor):
    if '+' in version_str:
        version_str = version_str.split('+')[0]
    version_arr = version_str.split('.')
    maj = int(version_arr[0])
    sub = int(version_arr[1])
    mi = int(version_arr[2])
    if maj >= major and sub >= subor and mi >= minor:
        return True
    return False


def get_op_nodes(graph):
    nodes = []
    for n in graph.nodes():
        op_node = OpNodeInfo()
        for i in n.inputs():
            op_node.inputs.append(i.debugName())
        for o in n.outputs():
            op_node.outputs.append(o.debugName())
        op_node.op = n.kind()
        op_node.scope = n.scopeName()
        nodes.append(op_node)
    return nodes


def find_adjoin_layer(src_node_name, may_be_dst_layers, dict_input, dict_output, src_node_must_be_layers=None):
    find_nodes = []
    for dst_layer in may_be_dst_layers:

        input_tensor = dst_layer.inputs[0]
        src_node = dict_output[input_tensor]
        if src_node != None and src_node.op == src_node_name:
            input_node_set = dict_input[input_tensor]
            if len(input_node_set) == 1 and dst_layer in input_node_set:
                if src_node_must_be_layers is None:
                    find_nodes.append((src_node, dst_layer))
                elif src_node in src_node_must_be_layers:
                    find_nodes.append((src_node, dst_layer))
    target_set = set([])
    src_set = set([])
    for s, d in find_nodes:
        src_set.add(s)
        target_set.add(d)
    return find_nodes, src_set, target_set


def find_adjoin_adjoin_layer(src_node_name, mid_node_name, may_be_dst_layers, dict_input, dict_output):
    find_nodes = []
    for dst_layer in may_be_dst_layers:
        dst_input_tensor = dst_layer.inputs[0]
        mid_node = dict_output[dst_input_tensor]
        if mid_node != None and mid_node.op == mid_node_name:
            mid_input_node_set = dict_input[dst_input_tensor]
            if len(mid_input_node_set) == 1 and dst_layer in mid_input_node_set:
                mid_input_tensor = mid_node.inputs[0]
                src_node = dict_output[mid_input_tensor]
                if src_node != None and src_node.op == src_node_name:
                    src_input_node_set = dict_input[mid_input_tensor]
                    if len(src_input_node_set) == 1 and mid_node in src_input_node_set:
                        find_nodes.append((src_node, mid_node, dst_layer))
    src_set = set([])
    mid_set = set([])
    dst_set = set([])
    for (s, m, d) in find_nodes:
        src_set.add(s)
        mid_set.add(m)
        dst_set.add(d)
    return find_nodes, src_set, mid_set, dst_set


def filter_layers(node_arr, op_name):
    list_node = []
    for n in node_arr:
        if n.op == op_name:
            list_node.append(n)
    return list_node


def parse_fuseable_conv_bn(node_arr, fused_bn=True, ahead_conv_relu=True, ahead_bn_relu=True, ahead_linear_relu=True, ahead_conv_sigmoid=True, ahead_linear_sigmoid=True):
    dict_output = {}
    for n in node_arr:
        for o in n.outputs:
            dict_output[o] = n
    dict_input = {}
    for n in node_arr:
        for i in n.inputs:
            if dict_input.get(i) == None:
                dict_input[i] = set([n])
            else:
                dict_input[i].add(n)
    list_bn = filter_layers(node_arr, 'aten::batch_norm')
    list_relu = filter_layers(node_arr, 'aten::relu')
    list_sigmoid = filter_layers(node_arr, 'aten::sigmoid')
    fused_conv_sigmoid = []
    fused_linear_sigmoid = []
    fused_linear_bias_sigmoid = []
    fused_conv_bn = []
    fused_conv_bn_relu = []
    fused_conv_relu = []
    fused_bn_relu = []
    fused_linear_relu = []
    fused_linear_bias_relu = []
    if fused_bn:
        fused_conv_bn, _, _ = find_adjoin_layer(
            'aten::_convolution', list_bn, dict_input, dict_output)
        if ahead_bn_relu:
            fused_conv_bn_relu, _, _, _ = find_adjoin_adjoin_layer(
                'aten::_convolution', 'aten::batch_norm', list_relu, dict_input, dict_output)
    if ahead_conv_relu:
        fused_conv_relu, _, _ = find_adjoin_layer(
            'aten::_convolution', list_relu, dict_input, dict_output)
    if ahead_conv_sigmoid:
        fused_conv_sigmoid, _, _ = find_adjoin_layer(
            'aten::_convolution', list_sigmoid, dict_input, dict_output)
    if ahead_linear_sigmoid:
        fused_linear_bias_sigmoid, _, _, _ = find_adjoin_adjoin_layer(
            'aten::matmul', 'aten::add_', list_sigmoid, dict_input, dict_output)
        fused_linear_sigmoid, _, _ = find_adjoin_layer(
            'aten::matmul', list_sigmoid, dict_input, dict_output)
    if ahead_bn_relu:
        fused_bn_relu, _, _ = find_adjoin_layer(
            'aten::batch_norm', list_relu, dict_input, dict_output)
    if ahead_linear_relu:
        fused_linear_bias_relu, _, _, _ = find_adjoin_adjoin_layer(
            'aten::matmul', 'aten::add_', list_relu, dict_input, dict_output)
        fused_linear_relu, _, _ = find_adjoin_layer(
            'aten::matmul', list_relu, dict_input, dict_output)

    return fused_conv_bn, fused_conv_bn_relu, fused_conv_relu, fused_bn_relu, fused_linear_relu, fused_linear_bias_relu, fused_conv_sigmoid, fused_linear_sigmoid, fused_linear_bias_sigmoid


def scope_to_module(root_module, scope):
    tail_name = OpNodeInfo.parse_scope_to_path(scope)
    module_arr_name = tail_name.split('.')
    module_cur = root_module
    module_cur_name = ''
    moduel_cur_father = root_module
    str_find = ''
    for sub_att_name in module_arr_name:
        str_find += sub_att_name+"."
        moduel_cur_father = module_cur
        module_cur = getattr(module_cur, sub_att_name)
        module_cur_name = sub_att_name
        assert module_cur is not None, 'can not find '+str_find
    return (moduel_cur_father, module_cur, module_cur_name)


def FuseConvBNAheadRelu(model, *args, fused_bn=True, ahead_conv_relu=True, ahead_bn_relu=True, ahead_linear_relu=True, ahead_conv_sigmoid=True, ahead_linear_sigmoid=True):
    SingletonConvFusedBnModules().clear()
    assert _check_version(torch.__version__, 1, 5,
                          0), 'error: torch version must greater than 1.5'
    graph = torch.jit.trace(model, *args)
    node_arr = get_op_nodes(graph.inlined_graph)
    fuseable_conv_bn, fuseable_conv_bn_relu, fuseable_conv_relu, fuseable_bn_relu, fuseable_linear_relu, fuseable_linear_bias_relu, fuseable_conv_sigmoid, fuseable_linear_sigmoid, fuseable_linear_bias_sigmoid = parse_fuseable_conv_bn(
        node_arr, fused_bn, ahead_conv_relu, ahead_bn_relu, ahead_linear_relu, ahead_conv_sigmoid, ahead_linear_sigmoid)
    module_paths = []
    if fused_bn:
        if ahead_bn_relu:
            for (conv, bn, _) in fuseable_conv_bn_relu:
                _, conv_module, _ = scope_to_module(model, conv.scope)
                setattr(conv_module, LINGER_AHEAD_RELU, True)
        for (conv, bn) in fuseable_conv_bn:
            conv_module_father, conv_module, conv_module_name = scope_to_module(
                model, conv.scope)
            bn_module_father, bn_module, bn_module_name = scope_to_module(
                model, bn.scope)
            if (type(conv_module) == torch.nn.Conv2d and type(bn_module) == torch.nn.BatchNorm2d) or \
                    (type(conv_module) == torch.nn.Conv1d and type(bn_module) == torch.nn.BatchNorm1d):
                fuseableconv_bn = FuseableConvBN(
                    conv_module_father, conv_module, bn_module_father, bn_module)
                SingletonConvFusedBnModules()._register(fuseableconv_bn)
                module_paths.append((conv.parse_scope(), bn.parse_scope()))
    if ahead_conv_relu:
        for (conv, _) in fuseable_conv_relu:
            _, conv_module, _ = scope_to_module(model, conv.scope)
            setattr(conv_module, LINGER_AHEAD_RELU, True)
    if ahead_bn_relu:
        for (bn, _) in fuseable_bn_relu:
            _, bn_module, _ = scope_to_module(model, bn.scope)
            setattr(bn_module, LINGER_AHEAD_RELU, True)
    if ahead_conv_sigmoid:
        for (conv, _) in fuseable_conv_sigmoid:
            _, conv_module, _ = scope_to_module(model, conv.scope)
            setattr(conv_module, LINGER_AHEAD_SIGMOID, True)
    if ahead_linear_sigmoid:
        for (linear, _) in fuseable_linear_sigmoid:
            _, linear_module, _ = scope_to_module(model, linear.scope)
            setattr(linear_module, LINGER_AHEAD_SIGMOID, True)
        for(linear, add, _) in fuseable_linear_bias_sigmoid:
            _, linear_module, _ = scope_to_module(model, linear.scope)
            setattr(linear_module, LINGER_AHEAD_SIGMOID, True)
    if ahead_linear_relu:
        for(linear, _) in fuseable_linear_relu:
            _, linear_module, _ = scope_to_module(model, linear.scope)
            setattr(linear_module, LINGER_AHEAD_RELU, True)
        for(linear, add, _) in fuseable_linear_bias_relu:
            _, linear_module, _ = scope_to_module(model, linear.scope)
            setattr(linear_module, LINGER_AHEAD_RELU, True)
    return module_paths


def FuseBNIntoConv(model, *args):
    r"""融合BN操作到Conv里

    Args:
        model(torch.nn.Module):模型
        *args:模型的trace位置参数
        **kwargs:模型trace的keyword参数
    Example:
        >>> net1 = shufflenet_v2_x1_0(pretrained=False)
        >>> net1.load_state_dict(torch.load(dict_file))
        >>> aa = net1(input)
        >>> linger.FuseBNIntoConv(net1,dummy_input)
        >>> net2 = linger.init(net1)
        >>> net2.load_state_dict(torch.load(dict_file))
    """
    assert False, 'FuseBNIntoConv is deprecated please use linger.trace_layers(root_net, trace_net, dummy_input)'


__all__ = ['FuseBNIntoConv', 'SingletonConvFusedBnModules',
           'EmptyBatchNorm', 'FuseConvBNAheadRelu']



import math

import lingerext
import numpy as np
import torch
import torch.nn.functional as F
import torch.onnx.symbolic_helper as sym_help

from linger.config import config
from linger.utils import PlatFormQuant, QuantMode, ScalerBuffer
epsilon = 1e-10


class Quant():
    @staticmethod
    def quant(x, bits=8, scale=-1, mode=QuantMode.QValue, *, quant_data='weight', ahead_relu=False, iq_zero_point=0):
        bound_value = None
        scale_local = None
        max_abs = None
        assert quant_data == 'input' or quant_data == 'output' or quant_data == 'weight' or quant_data == 'hidden' or quant_data == 'conv_output', 'invalid quant_data, please confirm'
        # assert drop_rate >= 0.0, 'invalid drop_rate param'
        zero_point = iq_zero_point
        if hasattr(x, 'zero_point'):
            zero_point = x.zero_point
        y = x.detach().clone()
        if (ahead_relu and quant_data == 'output') or (ahead_relu and quant_data == 'conv_output'):
            x = F.relu(x)
        if mode == QuantMode.QValue:
            max_abs = 0
            bound_value = math.pow(2, bits-1)-1
            if scale > 0:
                scale_local = scale
                max_abs = (bound_value+zero_point) / scale_local
            else:
                min_x = torch.min(x)
                max_x = torch.max(x)
                if min_x == max_x == 0:
                    scale_local = math.pow(2, bits)
                else:
                    max_abs = torch.max(-min_x, max_x)
                    max_value = round(
                        math.log((bound_value+zero_point) / max_abs, 2))
                    scale_local = math.pow(2, max_value)
                    max_abs = (bound_value+zero_point) / scale_local
        else:
            print('Error Quant Mode!!!')

        scale_local = ScalerBuffer(scale_local)
        x = y * scale_local
        
        #not iqtensor or test
        if config.PlatFormQuant.platform_quant == PlatFormQuant.luna_quant:
            x_quant = (x + 0.5).floor()
        else:
            assert False, "linger only support luna quant."
        x_quant.clamp_(-bound_value-1+zero_point, bound_value+zero_point)
        # drop_rate means quant percent, if drop_rate=1.0, all quant
        
        x = x_quant.float()
        return x, scale_local, max_abs

    @staticmethod
    def dequant(x, scale):
        scale_tensor = None
        if isinstance(scale, (float, np.float32)):
            scale_tensor = torch.tensor(
                scale, dtype=torch.float32, device=x.device)
        else:
            scale_tensor = torch.tensor(
                scale.data, dtype=torch.float32, device=x.device)
        return (x/scale_tensor).float()

    @staticmethod
    def running_to_scale(running_data, bits, mode=QuantMode.QValue, zero_point=0):
        running_data = ScalerBuffer(running_data)
        scale_data = None
        bound_value = math.pow(2, bits-1)-1
        if mode == QuantMode.QValue:
            max_value = round(math.log(
                (bound_value+zero_point)/running_data(), 2)) if running_data() != 0 else 0.0
            scale_data = math.pow(2, max_value)
        scale_data = ScalerBuffer(scale_data)
        return scale_data

class ClipGrad(torch.autograd.Function):
    @staticmethod
    def forward(ctx, data, clip_threshold):
        ctx.clip_threshold = clip_threshold
        return data

    @staticmethod
    def backward(ctx, grad_output):
        clip_threshold = ctx.clip_threshold
        grad_output = grad_output.clamp(-clip_threshold, clip_threshold)
        return grad_output, None


class NormalizeFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, data, bound, training=True, is_weight=True):
        data, temp_loc1, temp_loc2 = lingerext.normalize_function_forward(
            data, float(bound), training, is_weight)
        ctx.save_for_backward(temp_loc1, temp_loc2)
        return data

    @staticmethod
    def backward(ctx, grad_output):
        temp_loc1, temp_loc2 = ctx.saved_tensors
        grad_output = lingerext.normalize_function_backward(
            grad_output, temp_loc1, temp_loc2)

        return grad_output, None, None, None

    @staticmethod
    def symbolic(g, input, bound, training, is_weight=True):
        dtype = input.type().scalarType()
        if dtype is None:
            dtype = 6  # float
        else:
            dtype = sym_help.scalar_type_to_onnx.index(
                sym_help.cast_pytorch_to_onnx[dtype])
        min_val = g.op("Constant", value_t=torch.tensor(-bound,
                       dtype=sym_help.scalar_type_to_pytorch_type[dtype]))
        max_val = g.op("Constant", value_t=torch.tensor(
            bound, dtype=sym_help.scalar_type_to_pytorch_type[dtype]))
        return g.op("Clip", *[input, min_val, max_val])


def normalize_data_with_config(input, normalize_data):
    if normalize_data is not None:
        input_clamp_data = normalize_data
        input = input.clamp(-input_clamp_data, input_clamp_data)
    return input


def normalize_weight_with_config(weights, normalize_weight, training=False):
    if normalize_weight is not None:
        weights = NormalizeFunction.apply(weights, normalize_weight, training)
    return weights


def normalize_bias_with_config(bias, normalize_bias, training=False):
    if normalize_bias is not None:
        bias = NormalizeFunction.apply(bias, normalize_bias, training)
    return bias



import math

import torch
import torch.nn as nn
import torch.nn.functional as F
from ..config import config
from ..ops.ops import ModuleIntConfig
from ..utils import Dump, QuantMode, ScalerBuffer
from .iqtensor import (IQTensor, from_torch_tensor, platform_to_string,
                       quantlinear)
from .ops import ModuleIntConfig
from .requant import Requant
from ..quant import Quant


class ReLU6IntFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input, min_thresh, max_thresh, data_bits, training, momentum, running_x, eval_scale_x, mode, quant, is_not_from_iqtensor, prefix, dump, path):

        scale_x = None
        if training:
            ctx.save_for_backward(input)
            zero_point = input.zero_point if isinstance(input, IQTensor) else 0
            is_iq_tensor = True if isinstance(input, IQTensor) else False
            ctx.value = zero_point, is_iq_tensor
            if isinstance(input, IQTensor):
                q_input, _, max_value_x = quant.quant(input, data_bits, eval_scale_x, mode=QuantMode.QValue, quant_data='input',
                                                      iq_zero_point=input.zero_point)
                scale_x = eval_scale_x
            else:
                q_input, scale_x, max_value_x = quant.quant(
                    input, data_bits, mode=mode, quant_data='input')
                scale_x = scale_x
                running_x.mul_(1-momentum).add_(momentum*max_value_x)
            bound_value = math.pow(2, data_bits-1)-1
            max_thresh_int = round(6 * scale_x)
            max_thresh_int = bound_value if max_thresh_int > bound_value else max_thresh_int
            q_outputs = q_input.clamp(0, max_thresh_int)
            outputs = quant.dequant(q_outputs, scale_x)
            max_thresh.fill_(max_thresh_int)
            ctx.scale = scale_x, data_bits
        else:
            assert running_x > 0, 'invalid running_x <= 0, please fintune first'
            scale_x = None
            scale_x = ScalerBuffer(quant.running_to_scale(
                running_x, data_bits, mode=mode))
            if isinstance(input, IQTensor):
                scale_x = eval_scale_x
                q_input, _, _ = quant.quant(
                    input, data_bits, scale_x, mode=QuantMode.QValue, quant_data='input', iq_zero_point=input.zero_point)
            else:
                q_input, _, _ = quant.quant(
                    input, data_bits, scale_x, mode=mode, quant_data='input')
            bound_value = math.pow(2, data_bits-1)-1
            max_thresh_int = round(6 * scale_x)
            max_thresh_int = bound_value if max_thresh_int > bound_value else max_thresh_int
            q_outputs = q_input.clamp(0, max_thresh_int)
            outputs = quant.dequant(q_outputs, scale_x)
            max_thresh.fill_(max_thresh_int)
            if dump:
                name_list = ["input", "outputs", "q_input",
                             "q_outputs", "scale_x", "running_x"]
                attr_list = [input, outputs, q_input,
                             q_outputs, scale_x.data, running_x.data]
                Dump.dump_file(prefix, ".ReLU6Int.", zip(
                    name_list, attr_list), path)

        if isinstance(scale_x, float):
            return from_torch_tensor(outputs, scale_x, data_bits)
        elif isinstance(scale_x, torch.Tensor):
            return from_torch_tensor(outputs, scale_x.item(), data_bits)
        else:
            return from_torch_tensor(outputs, scale_x.data, data_bits)

    @staticmethod
    def backward(ctx, gradoutput):
        input, = ctx.saved_tensors
        scale_x, data_bits = ctx.scale
        zero_point, is_iq_tensor = ctx.value
        if is_iq_tensor:
            f_input = input.data
        else:
            q_input, _, _ = Quant.quant(
                input.data, data_bits, scale_x, mode=QuantMode.QValue, quant_data='input')
            f_input = Quant.dequant(q_input, scale_x)
        f_input = f_input.detach().clone().requires_grad_(True)
        gradInput = None
        with torch.enable_grad():
            y = F.hardtanh(f_input, 0, 6, False)
            gradInput, = torch.autograd.grad(y, f_input, gradoutput)
        return gradInput, None, None, None, None, None, None, None, None, None, None, None, None, None

    @staticmethod
    def symbolic(g, input, min_thresh, max_thresh, data_bits, training, momentum, running_x, scale_x, mode, quant, is_not_from_iqtensor, prefix, dump, path):
        op_inner = None
        if is_not_from_iqtensor:
            platform_quant = platform_to_string(
                config.PlatFormQuant.platform_quant)
            op_inner = quantlinear(g, input, scale_x(),
                                   platform_quant, data_bits)
        if is_not_from_iqtensor:
            input_list = [op_inner, min_thresh, max_thresh]
        else:
            input_list = [input, min_thresh, max_thresh]
        return g.op("Clip", *input_list)


class ReLU6Int(nn.ReLU6, ModuleIntConfig):
    r"""
        实现relu6的clamp按照定点形式计算，而非浮点6

    Args:
        data_bits(int): 输入量化位数
        mode(Enum):量化计算方式，支持MaxValue, QValue
        o_bits(int):输出量化位数 
        scale_x(np.float32): 输出量化scale，与输入scale_x一致
    """

    def __init__(self, data_bits=8, mode=QuantMode.QValue):
        nn.ReLU6.__init__(self)
        ModuleIntConfig.__init__(
            self, data_bits=data_bits, mode=mode, o_bits=None)
        self.momentum = 0.1
        self.is_not_from_iqtensor = True
        self.prefix = ""
        self.dump = False
        self.path = ""
        self.register_buffer('running_x', torch.zeros(1))
        self.register_buffer('scale_x', torch.zeros(1))
        self.register_buffer('min_thresh', torch.tensor(0, dtype=torch.int8))
        self.register_buffer('max_thresh', torch.tensor(0, dtype=torch.int8))

    def forward(self, input):
        scale_x = ScalerBuffer(self.scale_x)
        running_x = ScalerBuffer(self.running_x)
        assert self.data_bits == 8, 'relu6int only support 8bit'
        if isinstance(input, IQTensor):
            self.is_not_from_iqtensor = False
            if input.bits != self.data_bits:
                input = Requant.apply(
                    input, input.bits, input.scale_data, self.data_bits)
            scale_x = ScalerBuffer(input.scale_data)
            running_x = ScalerBuffer(input.running_data)
        output = ReLU6IntFunction.apply(input.contiguous(), self.min_thresh, self.max_thresh, self.data_bits, self.training, self.momentum,
                                        running_x, scale_x, self.quant_mode, self.quant, self.is_not_from_iqtensor, self.prefix, self.dump, self.path)
        self.running_x.fill_(running_x())
        self.scale_x.fill_(scale_x())

        return output

    def state_dict(self, destination=None, prefix='', keep_vars=False):
        return ModuleIntConfig.state_dict_global(self, destination, prefix, keep_vars)

    def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict,
                              missing_keys, unexpected_keys, error_msgs):
        ModuleIntConfig._load_from_state_dict_global(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import torch
import torch.nn as nn
import torch.nn.functional as F

from ..quant import NormalizeFunction


class NormalizeConvBN2d(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, kernel_size, stride=1, padding=0, dilation=1, groups=1, bias=True, padding_mode='zeros',
                 eps=1e-5, momentum=0.1, affine=True, track_running_stats=True,
                 normalize_data=None, normalize_weight=None, normalize_bias=None, ahead_relu=False) -> None:
        assert normalize_data is None or normalize_data > 0, 'normalize value is None or must >0'
        assert normalize_weight is None or normalize_weight > 0, 'normalize value is None or must >0'
        assert normalize_bias is None or normalize_bias > 0, 'normalize value is None or must >0'
        super(NormalizeConvBN2d, self).__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size,
                              stride, padding, dilation, groups, bias, padding_mode)
        self.bn = nn.BatchNorm2d(
            out_channels, eps, momentum, affine, track_running_stats)

        self.normalize_data = normalize_data
        self.normalize_weight = normalize_weight
        self.normalize_bias = normalize_bias
        self.ahead_relu = ahead_relu

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        if self.training:
            conv_rlt = self.conv(input)
            batchsize, channels, height, width = conv_rlt.size()
            numel = batchsize * height * width
            conv_rlt = conv_rlt.permute(
                1, 0, 2, 3).contiguous().view(channels, numel)
            sum_ = conv_rlt.sum(1)
            sum_of_square = conv_rlt.pow(2).sum(1)
            mean = sum_ / numel
            sumvar = sum_of_square - sum_ * mean
            unbias_var = sumvar / (numel - 1)
            unbias_var = torch.clamp(unbias_var, min=0.0)
            self.bn.running_mean = (
                (1 - self.bn.momentum) * self.bn.running_mean + self.bn.momentum * mean.detach())
            self.bn.running_var = (
                (1 - self.bn.momentum) * self.bn.running_var + self.bn.momentum * unbias_var.detach())

            bias_var = sumvar / numel
            bias_var = torch.clamp(bias_var, min=0.0)
            inv_std = 1 / (bias_var + self.bn.eps).pow(0.5)
            bn_rlt = ((conv_rlt - mean.unsqueeze(1)) * inv_std.unsqueeze(1)
                      * self.bn.weight.unsqueeze(1) + self.bn.bias.unsqueeze(1))
            bn_rlt = bn_rlt.view(channels, batchsize, height, width).permute(
                1, 0, 2, 3).contiguous()

            w_bn = self.bn.weight.div(torch.sqrt(self.bn.eps + unbias_var))
            new_weight = self.conv.weight.mul(w_bn.view(-1, 1, 1, 1))
            if self.conv.bias is not None:
                b_conv = self.conv.bias
            else:
                b_conv = torch.zeros(
                    self.conv.weight.size(0), device=input.device)
            b_bn = self.bn.bias - \
                self.bn.weight.mul(mean).div(
                    torch.sqrt(unbias_var + self.bn.eps))
            new_bias = b_conv.mul(w_bn) + b_bn
            alpha = 1.0
            if self.normalize_weight is not None:
                new_weight = NormalizeFunction.apply(
                    new_weight, self.normalize_weight, self.training)
            if self.normalize_bias is not None:
                new_bias = NormalizeFunction.apply(
                    new_bias, self.normalize_bias, self.training)
            new_conv_rlt = F.conv2d(input, new_weight, new_bias, self.conv.stride,
                                    self.conv.padding, self.conv.dilation, self.conv.groups)
            # out = new_conv_rlt
            out = alpha * bn_rlt + (1 - alpha) * new_conv_rlt
        else:
            w_bn = self.bn.weight.div(torch.sqrt(
                self.bn.eps + self.bn.running_var))
            new_weight = self.conv.weight.mul(w_bn.view(-1, 1, 1, 1))
            if self.conv.bias is not None:
                b_conv = self.conv.bias
            else:
                b_conv = torch.zeros(
                    self.conv.weight.size(0), device=input.device)
            b_bn = self.bn.bias - self.bn.weight.mul(self.bn.running_mean).div(
                torch.sqrt(self.bn.running_var + self.bn.eps))
            new_bias = b_conv.mul(w_bn) + b_bn
            if self.normalize_weight is not None:
                new_weight = NormalizeFunction.apply(
                    new_weight, self.normalize_weight, self.training)
            if self.normalize_bias is not None:
                new_bias = NormalizeFunction.apply(
                    new_bias, self.normalize_bias, self.training)
            out = F.conv2d(input, new_weight, new_bias, self.conv.stride,
                              self.conv.padding, self.conv.dilation, self.conv.groups)

        if self.normalize_data is not None:
            out = NormalizeFunction.apply(out, self.normalize_data, self.training, False)
        return out

    def extra_repr(self):
        return 'normalize_data:{normalize_data},normalize_weight:{normalize_weight},normalize_bias:{normalize_bias},ahead_relu:{ahead_relu}'.format(**self.__dict__)


__all__ = ['NormalizeConvBN2d']



import torch
import torch.nn as nn
import torch.nn.functional as F

from ..config import config
from ..quant import (Quant, normalize_bias_with_config,
                     normalize_data_with_config, normalize_weight_with_config)
from ..utils import Dump, PlatFormQuant, QuantMode, ScalerBuffer
from .iqtensor import (IQTensor, from_torch_tensor, platform_to_string,
                       quantlinear)
from .ops import ModuleIntConfig
from .requant import Requant


class Conv1dFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input, weights, bias, kernel_size, padding, stride, dilation, groups, params,
                data_bits, parameter_bits, training, momentum, running_x, running_w, running_o, eval_scale_x, eval_scale_w, eval_scale_o,
                prefix, dump, path, mode, o_bits, quant, ahead_relu, is_not_from_iqtensor, clamp_data, clamp_weight, clamp_bias):
        scale_o = None
        if training:
            ctx.clamp_data = clamp_data
            ctx.clamp_weight = clamp_weight
            ctx.clamp_bias = clamp_bias
            zero_point = input.zero_point if isinstance(input, IQTensor) else 0
            is_iq_tensor = True if isinstance(input, IQTensor) else False
            ctx.value = zero_point, is_iq_tensor
            ctx.bits = data_bits, parameter_bits, o_bits
            saved_tensors = [input, weights, bias, params]
            if isinstance(input, IQTensor):
                q_input, _, max_value_x = quant.quant(
                    input.data, data_bits, eval_scale_x, mode=QuantMode.QValue, quant_data='input', iq_zero_point=input.zero_point)
                scale_x = eval_scale_x
            else:
                q_input, scale_x, max_value_x = quant.quant(
                    input.data, data_bits, mode=mode, quant_data='input')
                running_x.mul_(1-momentum).add_(momentum*max_value_x)

            # weights = normalize_weight_with_config(weights, clamp_weight, False)
            # if bias is not None:
            #     bias = normalize_bias_with_config(bias, clamp_bias, False)
            q_weights, scale_w, max_value_w = quant.quant(
                weights, parameter_bits, mode=mode, quant_data='weight')
            running_w.mul_(1-momentum).add_(momentum*max_value_w)
            q_input = q_input.float() if data_bits + parameter_bits <= 16 else q_input.double()
            q_weights = q_weights.float() if data_bits + \
                parameter_bits <= 16 else q_weights.double()
            q_outputs = F.conv1d(q_input, q_weights, stride=stride,
                                 padding=padding, dilation=dilation, groups=groups)
            outputs = None
            q_bias = None

            if config.PlatFormQuant.platform_quant in (PlatFormQuant.luna_quant,):
                assert mode == QuantMode.QValue, 'castor quant only support QValue and o_bits=None'
                if bias is not None:
                    q_bias = (bias * scale_w * scale_x + 0.5).floor()
                    if data_bits + parameter_bits <= 16:
                        q_bias = q_bias.float()
                    else:
                        q_bias = q_bias.double()
                    q_outputs += q_bias.reshape(1, -1, 1)
                outputs = quant.dequant(q_outputs, scale_x*scale_w)
            else:
                assert False, "linger only support luna quant."
            # ctx.save_for_backward(*saved_tensors)
            # f_input = quant.dequant(q_input, scale_x)
            # f_weights = quant.dequant(q_weights, scale_w)
            # f_bias = None if bias is None else quant.dequant(q_bias, scale_x*scale_w)
            # saved_tensors = [f_input, f_weights, f_bias, params]

            out_tensor = outputs
            if o_bits is not None:
                outputs = normalize_data_with_config(outputs, clamp_data)
                out_tensor = outputs
                q_outputs, scale_o, max_value_o = quant.quant(
                    outputs, o_bits, mode=mode, quant_data='output', ahead_relu=ahead_relu)
                outputs = quant.dequant(q_outputs, scale_o)
                running_o.mul_(1-momentum).add_(momentum*max_value_o)
            ctx.scale = scale_x, scale_w, scale_o
            saved_tensors += [out_tensor]
            ctx.save_for_backward(*saved_tensors)
        else:
            assert running_x > 0, 'invalid running_x, please finetune training before eval'
            scale_x = None
            scale_w = None
            scale_o = eval_scale_o
            if weights.dtype == torch.float32:
                scale_x = ScalerBuffer(quant.running_to_scale(
                    running_x, data_bits, mode=mode))
                q_weights, scale_w, _ = quant.quant(
                    weights, parameter_bits, mode=mode, quant_data='weight')
                scale_w = ScalerBuffer(scale_w)
                if o_bits is not None:
                    assert running_o > 0, 'invalid running_o <= 0, please finetune training'
                    scale_o = ScalerBuffer(quant.running_to_scale(
                        running_o, o_bits, mode=mode))
            else:
                scale_x = eval_scale_x
                scale_w = eval_scale_w
                q_weights = weights.double()
                if o_bits is not None:
                    scale_o = eval_scale_o

            if isinstance(input, IQTensor):
                scale_x = eval_scale_x
                q_input, _, _ = quant.quant(
                    input.data, data_bits, scale_x, mode=QuantMode.QValue, quant_data='input', iq_zero_point=input.zero_point)
            else:
                q_input, scale_x, _ = quant.quant(
                    input.data, data_bits, scale_x, mode=mode, quant_data='input')
            q_input = q_input.double()
            q_weights = q_weights.double()
            q_outputs = F.conv1d(q_input, q_weights, stride=stride,
                                 padding=padding, dilation=dilation, groups=groups)
            # # ensure bias clamp
            # if bias is not None and bias.dtype == torch.float32:
            #     bias = normalize_bias_with_config(bias, clamp_bias, False)
            outputs = None
            q_bias = None
            if config.PlatFormQuant.platform_quant in (PlatFormQuant.luna_quant,):
                assert mode == QuantMode.QValue, 'castor quant only support QValue and o_bits=None'
                if bias is not None:
                    if bias.dtype == torch.float32:
                        q_bias = (bias * scale_w * scale_x + 0.5).floor()
                        if data_bits + parameter_bits <= 16:
                            q_bias = q_bias.float().double()
                        else:
                            q_bias = q_bias.double()
                    else:
                        q_bias = bias.double()
                    q_outputs += q_bias.reshape(1, -1, 1)
                outputs = quant.dequant(q_outputs, scale_x*scale_w)
            else:
                assert False, "linger only support luna quant."
            if o_bits is not None:
                if config.PlatFormQuant.platform_quant == PlatFormQuant.luna_quant:
                    q_outputs, _, _ = quant.quant(
                        outputs, o_bits, scale_o, mode=mode, quant_data='output')
                outputs = quant.dequant(q_outputs, scale_o)
            if dump:
                if q_bias is not None:
                    name_list = ["input", "weights", "bias", "outputs", "q_input",  "q_weights", "q_bias",
                                 "q_outputs", "scale_x", "scale_w", "scale_o", "running_x", "running_w", "running_o"]
                    attr_list = [input, weights, bias, outputs, q_input, q_weights, q_bias, q_outputs,
                                 scale_x.data, scale_w.data, scale_o.data, running_x.data, running_w.data, running_o.data]
                    Dump.dump_file(prefix, ".Conv1dInt.",
                                   zip(name_list, attr_list), path)
                else:
                    name_list = ["input", "weights", "outputs", "q_input",  "q_weights", "q_outputs",
                                 "scale_x", "scale_w", "scale_o", "running_x", "running_w", "running_o"]
                    attr_list = [input, weights, outputs, q_input, q_weights, q_outputs, scale_x.data,
                                 scale_w.data, scale_o.data, running_x.data, running_w.data, running_o.data]
                    Dump.dump_file(prefix, ".Conv1dInt.",
                                   zip(name_list, attr_list), path)
        if o_bits is None:
            return outputs
        elif isinstance(scale_o, float):
            return from_torch_tensor(outputs, scale_o, o_bits)
        elif isinstance(scale_o, torch.Tensor):
            return from_torch_tensor(outputs, scale_o.item(), o_bits)
        else:
            return from_torch_tensor(outputs, scale_o.data, o_bits)

    @staticmethod
    def backward(ctx, gradOutput):
        clamp_data = ctx.clamp_data
        data_bits, parameter_bits, o_bits = ctx.bits
        zero_point, is_iq_tensor = ctx.value
        scale_x, scale_w, scale_o = ctx.scale
        input, weights, bias, params, outputs = ctx.saved_tensors
        # zero_point = input.zero_point if isinstance(input, IQTensor) else 0
        if is_iq_tensor:
            f_input = input.data
        else:
            q_input, _, _ = Quant.quant(
                input.data, data_bits, scale_x, mode=QuantMode.QValue, quant_data='input')
            f_input = Quant.dequant(q_input, scale_x)
        f_input = f_input.detach().clone().requires_grad_(True)
        q_weights, _, _ = Quant.quant(
            weights.data, parameter_bits, scale_w, mode=QuantMode.QValue, quant_data='weight')
        f_weights = Quant.dequant(q_weights, scale_w)
        f_weights = f_weights.detach().clone().requires_grad_(True)
        bias = None if bias is None else bias.detach().clone().requires_grad_(True)

        stride = int(params[0])
        padding = int(params[1])
        dilation = int(params[2])
        groups = int(params[3])
        gradBias = None
        # input_clamp = input
        with torch.enable_grad():
            z = F.conv1d(f_input, f_weights, bias,
                         stride, padding, dilation, groups)
            if o_bits is not None:
                z = normalize_data_with_config(z, clamp_data)
            if bias is not None:
                gradInput, gradWeight, gradBias = torch.autograd.grad(
                    z, (f_input, f_weights, bias), gradOutput)
            else:
                gradInput, gradWeight, = torch.autograd.grad(
                    z, (f_input, f_weights), gradOutput)

        return gradInput, gradWeight, gradBias, None, None, None, None, None, None, None, None, None, None, None, None,\
            None, None, None, None, None, None, None, None, None, None, None, None, None, None,\
            None, None, None, None, None, None

    @staticmethod
    def symbolic(g, input, weights, bias, kernel_size, padding, stride, dilation, groups, params,
                 data_bits, parameter_bits, training, momentum, running_x, running_w, running_o, scale_x, scale_w, scale_o,
                 prefix, dump, path, mode, o_bits, quant, ahead_relu, is_not_from_iqtensor, clamp_data, clamp_weight, clamp_bias):
        op_inner = None
        if is_not_from_iqtensor:
            platform_quant = platform_to_string(
                config.PlatFormQuant.platform_quant)
            op_inner = quantlinear(g, input, scale_x(),
                                   platform_quant, data_bits)
        paddings = padding + padding
        param_dict = {'scale_x_f': scale_x(), 'scale_w_f': scale_w(),
                      'dilations_i': dilation, 'group_i': groups, 'kernel_shape_i': kernel_size, 'pads_i': paddings, 'strides_i': stride,
                      'data_bits_i': data_bits, 'parameter_bits_i': parameter_bits}
        if is_not_from_iqtensor:
            input_list = [op_inner, weights]
        else:
            input_list = [input, weights]
        if bias is not None:
            input_list.append(bias)
        if o_bits is not None:
            param_dict['scale_o_f'] = scale_o()
            param_dict['o_bits_i'] = o_bits

        platform_quant = platform_to_string(
            config.PlatFormQuant.platform_quant)
        param_dict['platform_quant_s'] = platform_quant

        return g.op("thinker::Conv1dInt", *input_list, **param_dict)


class Conv1dInt(nn.Conv1d, ModuleIntConfig):
    r"""实现Conv1dInt的量化训练与测试，继承自nn.Conv1d,

    Args:
        in_channels out_channels kernel_size stride padding dilation groups bias padding_mode
        与nn.Conv2d一致的参数
        data_bits(int): 输入量化位数
        parameter_bits(int): 参数量化位数
        mode(Enum): 量化方式，支持MaxValue与Qvalue
        o_bits(int, default=None):输出量化位数
        clamp_data(float or None): 针对输出的clamp数值
        clamp_weight(float or None): 针对转为weight的clamp数值
        clamp_bias(float or None): 与clamp_weight一致
        ahead_relu(bool): 是否做融合relu之后的数值统计scale

    """

    def __init__(self, in_channels, out_channels, kernel_size, stride=1,
                 padding=0, dilation=1, groups=1, bias=True, padding_mode='zeros',
                 data_bits=8, parameter_bits=8, mode=QuantMode.QValue, o_bits=None,
                 clamp_data=None, clamp_weight=None, clamp_bias=None, ahead_relu=False):
        nn.Conv1d.__init__(self, in_channels, out_channels,
                           kernel_size, stride, padding, dilation, groups, bias)
        ModuleIntConfig.__init__(
            self, data_bits=data_bits, parameter_bits=parameter_bits, mode=mode, o_bits=o_bits)
        self.params = torch.Tensor(
            [self.stride[0], self.padding[0], self.dilation[0], groups])
        self.momentum = 0.1
        self.prefix = ""
        self.dump = False
        self.path = ""
        self.ahead_relu = ahead_relu
        self.is_not_from_iqtensor = True
        self.clamp_data = clamp_data
        self.clamp_weight = clamp_weight
        self.clamp_bias = clamp_bias
        self.register_buffer('running_x', torch.zeros(1))
        self.register_buffer('running_w', torch.zeros(1))
        self.register_buffer('running_o', torch.zeros(1))
        self.register_buffer('scale_x', torch.zeros(1))
        self.register_buffer('scale_w', torch.zeros(1))
        self.register_buffer('scale_o', torch.zeros(1))

    def forward(self, input):
        scale_x = ScalerBuffer(self.scale_x)
        running_x = ScalerBuffer(self.running_x)
        if isinstance(input, IQTensor):
            self.is_not_from_iqtensor = False
            if input.bits != self.data_bits:
                input = Requant.apply(
                    input, input.bits, input.scale_data, self.data_bits)
            scale_x = ScalerBuffer(input.scale_data)
            running_x = ScalerBuffer(input.running_data)
        running_w = ScalerBuffer(self.running_w)
        running_o = ScalerBuffer(self.running_o)
        scale_w = ScalerBuffer(self.scale_w)
        scale_o = ScalerBuffer(self.scale_o)
        weight = self.weight
        bias = self.bias
        if weight.dtype == torch.float32:
            weight = normalize_weight_with_config(
                self.weight, self.clamp_weight, self.training)
            if self.bias is not None:
                bias = normalize_bias_with_config(
                    self.bias, self.clamp_bias, self.training)
        ret = Conv1dFunction.apply(input, weight, bias,
                                   self.kernel_size, self.padding, self.stride, self.dilation, self.groups,
                                   self.params, self.data_bits, self.parameter_bits, self.training, self.momentum,
                                   running_x, running_w, running_o, scale_x, scale_w, scale_o,
                                   self.prefix, self.dump, self.path, self.quant_mode, self.o_bits, self.quant, self.ahead_relu, self.is_not_from_iqtensor,
                                   self.clamp_data, self.clamp_weight, self.clamp_bias)

        self.running_x.fill_(running_x())
        self.running_w.fill_(running_w())
        self.running_o.fill_(running_o())
        self.scale_x.fill_(scale_x())
        self.scale_w.fill_(scale_w())
        self.scale_o.fill_(scale_o())
        return ret

    def state_dict(self, destination=None, prefix='', keep_vars=False):
        return ModuleIntConfig.state_dict_global(self, destination, prefix, keep_vars)

    def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict,
                              missing_keys, unexpected_keys, error_msgs):
        ModuleIntConfig._load_from_state_dict_global(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)

    def extra_repr(self):
        s = nn.Conv1d.extra_repr(self)
        extra_s = ',clamp_data:{clamp_data},clamp_weight:{clamp_weight},clamp_bias:{clamp_bias},ahead_relu:{ahead_relu}'.format(
            **self.__dict__)
        extra_s += ',data_bits:{data_bits},parameter_bits:{parameter_bits},o_bits:{o_bits}'.format(
            **self.__dict__)
        return s+extra_s



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import linger
import torch
import torch.nn as nn
import numpy

def test_lstmpint_net():

    def getacc(lprob, target):
        num_class = lprob.size()[1]
        _, new_target = torch.broadcast_tensors(lprob, target)

        remove_pad_mask = new_target.ne(-1)
        lprob = lprob[remove_pad_mask]

        target = target[target!=-1]
        target = target.unsqueeze(-1)


        lprob = lprob.reshape((-1, num_class))

        preds = torch.argmax(lprob, dim=1)
        
        correct_holder = torch.eq(preds.squeeze(), target.squeeze()).float()

        num_corr = correct_holder.sum()
        num_sample = torch.numel(correct_holder)
        acc = num_corr/num_sample
        return acc


    class Net(nn.Module):
        def __init__(self):
            super(Net, self).__init__()
            self.conv0 = nn.Conv2d(1, 100, kernel_size=(1,3), padding=(0,1), groups=1, bias=True)
            self.bn0 = nn.BatchNorm2d(100)
            self.relu0 = nn.ReLU()
            self.conv1 = nn.Conv2d(100, 100, kernel_size=(1,3), padding=(0,1), groups=1, bias=True)
            self.bn1 = nn.BatchNorm2d(100)
            self.relu1 = nn.ReLU()
            self.lstmp = nn.LSTM(100, 100, num_layers=1, batch_first=True, bidirectional=False)
            # self.lstmp = nn.LSTM(100, 50, num_layers=1, batch_first=True, bidirectional=True)
            self.final_conv = nn.Conv2d(100, 10, 1, 1, 0)
        def forward(self, input, batch_lengths=None, initial_state=None):
            x = self.conv0(input)
            x = self.bn0(x)
            x = self.relu0(x)
            x = self.conv1(x)
            x = self.bn1(x)
            x = self.relu1(x)
            n, c, h, w = x.shape
            x = x.reshape(n, -1, 1, w).squeeze(2)
            x = x.permute(0, 2, 1)  #b t d
            # !! !!! !! !!! !! !! !! !!!! !! !! ! !!!! !!! !
            # 此处之前和之后的pack_padded_sequence和pad_packed_sequence 需要写全，不能用from torch.nn.utils.rnn import pack_padded_sequence,pad_packed_sequence 使用
            # 直接写全torch.nn.utils.rnn.pack_padded_sequence(_,_,_,_)   torch.nn.utils.rnn.pad_packed_sequence(_,_,_,_)
            # 不然linger替换不了函数指针    会导致运行出错
            x = nn.utils.rnn.pack_padded_sequence(x, batch_lengths, batch_first=True, enforce_sorted=False)
            x, hidden = self.lstmp(x, initial_state) #output b, t, h (10, 10, 100)
            x, _ = nn.utils.rnn.pad_packed_sequence(x, batch_first=True)
            x = x.permute(2, 0, 1)
            d, b, t = x.shape
            x = x.reshape((1, d, 1, b*t)) # (1, 100, 1, 100)
            x = self.final_conv(x)  #(1, 10, 1, 100) (d, b*t)
            return x
    torch.manual_seed(1)
    torch.cuda.manual_seed_all(1)
    numpy.random.seed(1)
    aa = torch.randn(10, 1, 1, 10).cuda()
    label = torch.randint(10, (10, 10)).cuda()   #class=10
    mask = torch.ones(10, 10)
    for i in range(9):
        index = numpy.random.randint(5, 10)
        mask[i, index:] = 0
        label[i, index:] = -1

    input_lengths = mask.long().sum(1).cpu().numpy()
    input_lengths = torch.tensor(input_lengths)#.cuda()
    print('input_lengths: ', input_lengths)
    # input_lengths = None
    # label = label.permute((1, 0))
    # batch_size = 10; hidden_size=50; size=2
    batch_size = 10; hidden_size=100; size=1
    initial_state = (torch.zeros(size, batch_size, hidden_size).cuda(), 
                                torch.zeros(size, batch_size, hidden_size).cuda())
    # initial_state = None
    net = Net().cuda()
    replace_modules = (nn.Conv2d, nn.LSTM, nn.BatchNorm2d)
    # replace_modules = (nn.LSTM,)
    linger.SetPlatFormQuant(platform_quant=linger.PlatFormQuant.luna_quant)
    net = linger.init(net, quant_modules=replace_modules)
    criterion = nn.CrossEntropyLoss(ignore_index = -1)

    optimizer = torch.optim.Adam(net.parameters(), lr = 0.001)
    loss = None
    for i in range(50):
        optimizer.zero_grad()
        out = net(aa, input_lengths, initial_state)
        out = out.squeeze().permute((1,0)) #(b*t, d)
        loss = criterion(out, label.reshape(-1))
        if i % 50 == 0:
            print('loss: ', loss)
            acc = getacc(out, label.reshape(-1, 1))
            print('train acc: ', acc)
        loss.backward()
        optimizer.step()

    net.eval()
    out1 = net(aa, input_lengths, initial_state)
    out1 = out1.squeeze().permute((1,0))
    acc = getacc(out1, label.reshape(-1, 1))
    print('test acc: ', acc)
    assert acc > 0.4
    input_lengths = torch.tensor(input_lengths)#.cuda()
    # net = torch.jit.trace(net, (aa, input_lengths, initial_state))
    print(net)
    with torch.no_grad():
        net.eval()
        torch.onnx.export(net, (aa, input_lengths, initial_state), "data.ignore/single_lstm_4.onnx",export_params=True,opset_version=12, operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    out2 = net(aa, input_lengths, initial_state)
    out2 = out2.squeeze().permute((1,0))
    assert (out1 == out2).all()